from eth_typing import Address

__all__ = ['CHAIN_ID', 'MAX_GAS_LIMIT', 'GAS_PER_ORDER', 'OrderBookContractAddress', 'IOCBookContractAddress',
           'ClearingHouseContractAddress', 'min_quantity', 'price_precision', 'HTTP_PROTOCOL', 'WS_PROTOCOL']

OrderBookContractAddress = Address("0x0300000000000000000000000000000000000000")
IOCBookContractAddress = Address("0x635c5F96989a4226953FE6361f12B96c5d50289b")
ClearingHouseContractAddress = Address("0x0300000000000000000000000000000000000002")

CHAIN_ID = 321123
MAX_GAS_LIMIT = 7_000_000  # 7 million
GAS_PER_ORDER = 300_000  # 300k

min_quantity = {
    0: 0.01,
    1: 0.1,
    2: 0.001,
    3: 0.1,
    4: 0.1,
    5: 0.1,
    6: 0.1,
    7: 1,
    8: 0.01,
    9: 0.1,
}

price_precision = {
    0: 6,
    1: 6,
    2: 6,
    3: 6,
    4: 6,
    5: 6,
    6: 6,
    7: 6,
    8: 6,
    9: 6,
}

HTTP_PROTOCOL = "https"
WS_PROTOCOL = "wss"



import asyncio
import functools
import random
import time
from decimal import Decimal, getcontext
from functools import wraps

from eth_account import Account
from eth_typing import Address
from eth_utils.hexadecimal import add_0x_prefix as add_0x_prefix_eth_utils
from hexbytes import HexBytes

from hubble_exchange.constants import get_minimum_quantity, get_price_precision

# Set the precision higher than default to handle larger numbers without issues.
getcontext().prec = 50


def float_to_scaled_int(val: float, scale: int) -> int:
    scaled_value = Decimal(str(val)) * (Decimal('10') ** Decimal(scale))
    return int(scaled_value)


def int_to_scaled_float(val: int, scale: int) -> float:
    return float(Decimal(val) / (Decimal('10') ** Decimal(scale)))


def get_address_from_private_key(private_key: str) -> Address:
    account = Account.from_key(private_key)
    return account.address

def get_new_salt() -> int:
    return int(str(time.time_ns()) + str(random.randint(0, 10000)))


def add_0x_prefix(value: str) -> HexBytes:
    return HexBytes(add_0x_prefix_eth_utils(value))


def get_precision(value: float) -> int:
    # Convert float to string
    value_str = str(value)

    # Check if there's a decimal point in the string representation
    if '.' not in value_str:
        return 0

    # Split at the decimal point and return the length of the second part
    return len(value_str.split('.')[1])


def validate_limit_order_like(order):
    if order.amm_index is None:
        raise ValueError("Order AMM index is not set")
    if order.base_asset_quantity is None:
        raise ValueError("Order base asset quantity is not set")
    if order.price is None:
        raise ValueError("Order price is not set")
    if order.reduce_only is None:
        raise ValueError("Order reduce only is not set")

    min_quantity = get_minimum_quantity(order.amm_index)
    scaled_min_quantity = float_to_scaled_int(min_quantity, 18)
    if not is_multiple(order.base_asset_quantity, scaled_min_quantity):
        unscaled_order_size = int_to_scaled_float(order.base_asset_quantity, 18)
        raise ValueError(f"Order quantity {unscaled_order_size} is not a multiple of minimum quantity {min_quantity}")

    allowed_price_precision = get_price_precision(order.amm_index)
    price = int_to_scaled_float(order.price, 6)
    if get_precision(price) > allowed_price_precision:
        raise ValueError(f"Decimal precision of order price {price} is greater than allowed precision {allowed_price_precision}")


def is_multiple(a, b, tolerance=1e-9):
    remainder = a % b
    return abs(remainder) < tolerance or abs(remainder - b) < tolerance


def async_ttl_cache(ttl=300):
    def decorator(func):
        cache = {}
        cache_expiry = {}

        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            cache_key = str(args) + str(kwargs)
            current_time = time.time()

            if cache_key in cache and (current_time - cache_expiry[cache_key] < ttl):
                return cache[cache_key]

            result = await func(*args, **kwargs)
            cache[cache_key] = result
            cache_expiry[cache_key] = current_time

            return result

        return wrapped

    return decorator


def timeit(method):
    @wraps(method)
    async def timed_async(*args, **kw):
        start_time = time.time()
        result = await method(*args, **kw)
        end_time = time.time()
        print(f"Async method {method.__name__} took {end_time - start_time:.4f} seconds")
        return result

    def timed_sync(*args, **kw):
        start_time = time.time()
        result = method(*args, **kw)
        end_time = time.time()
        print(f"Sync method {method.__name__} took {end_time - start_time:.4f} seconds")
        return result

    if asyncio.iscoroutinefunction(method):
        return timed_async
    else:
        return timed_sync



from eth_typing import Address
from web3 import Web3

__all__ = ['CHAIN_ID', 'MAX_GAS_LIMIT', 'GAS_PER_ORDER', 'OrderBookContractAddress', 'IOCBookContractAddress',
           'ClearingHouseContractAddress', 'min_quantity', 'price_precision', 'HTTP_PROTOCOL', 'WS_PROTOCOL']


OrderBookContractAddress = Web3.to_checksum_address(Address("0x03000000000000000000000000000000000000b0"))
LimitOrderBookContractAddress = Web3.to_checksum_address(Address("0x03000000000000000000000000000000000000b3"))
IOCBookContractAddress = Web3.to_checksum_address(Address("0x03000000000000000000000000000000000000b4"))
SignedOrderBookContractAddress = Web3.to_checksum_address(Address("0x36C02dA8a0983159322a80FFE9F24b1acfF8B570"))
ClearingHouseContractAddress = Web3.to_checksum_address(Address("0x03000000000000000000000000000000000000b2"))

CHAIN_ID = 321123
MAX_GAS_LIMIT = 7_000_000  # 7 million
GAS_PER_ORDER = 300_000  # 300k

min_quantity = {
    0: 0.01,
}

price_precision = {
    0: 6,
}

HTTP_PROTOCOL = "http"
WS_PROTOCOL = "ws"

allowed_candle_intervals = ["1m", "3m", "5m", "15m", "30m", "1h", "4h", "8h", "1d", "1w", "1M"]




import asyncio
import os
from typing import Awaitable, Callable, Dict, List

from eth_typing import Address
from hexbytes import HexBytes
from web3 import AsyncHTTPProvider, AsyncWeb3, HTTPProvider, Web3
from web3.datastructures import AttributeDict
from web3.eth.async_eth import AsyncEth
from web3.exceptions import TimeExhausted
from web3.main import Web3, get_async_default_modules
from web3.method import Method, default_root_munger
from web3.middleware import (async_geth_poa_middleware,
                             construct_simple_cache_middleware,
                             geth_poa_middleware)
from web3.middleware.async_cache import _async_simple_cache_middleware
from web3.types import RPCEndpoint, _Hash32

from hubble_exchange.errors import OrderNotFound, TraderNotFound
from hubble_exchange.models import (GetPositionsResponse, OpenOrder,
                                    OrderBookDepthResponse,
                                    OrderStatusResponse)
from hubble_exchange.utils import int_to_scaled_float

sync_web3_client = None
async_web3_client = None


class HubblenetEth(AsyncEth):
    _get_transaction_status: Method[Callable[[_Hash32], Awaitable[Dict]]] = Method(RPCEndpoint("eth_getTransactionStatus"), mungers=[default_root_munger])
    _get_order_status: Method[Callable[[_Hash32], Awaitable[Dict]]] = Method(RPCEndpoint("trading_getOrderStatus"), mungers=[default_root_munger])
    _get_margin_and_positions: Method[Callable[[Address], Awaitable[Dict]]] = Method(RPCEndpoint("trading_getMarginAndPositions"), mungers=[default_root_munger])
    _get_order_book_depth: Method[Callable[[int], Awaitable[Dict]]] = Method(RPCEndpoint("trading_getTradingOrderBookDepth"), mungers=[default_root_munger])
    _get_open_orders: Method[Callable[[Address, str], Awaitable[Dict]]] = Method(RPCEndpoint("orderbook_getOpenOrders"), mungers=[default_root_munger])
    _place_signed_orders: Method[Callable[[str], Awaitable[Dict]]] = Method(RPCEndpoint("order_placeSignedOrders"), mungers=[default_root_munger])
    _get_clearing_house_vars: Method[Callable[[Address], Awaitable[Dict]]] = Method(RPCEndpoint("testing_getClearingHouseVars"), mungers=[default_root_munger])

    async def get_order_status(self, order_id: _Hash32) -> OrderStatusResponse:
        try:
            response = await self._get_order_status(order_id)
            return OrderStatusResponse(**response)
        except ValueError as e:
            if len(e.args) > 0 and e.args[0].get('message', '') == "order not found":
                raise OrderNotFound()
            else:
                raise e

    async def get_margin_and_positions(self, trader: Address) -> GetPositionsResponse:
        try:
            margin_response, clearing_house_vars = await asyncio.gather(
                self._get_margin_and_positions(trader),
                self._get_clearing_house_vars(trader)
            )

            margin_response = dict(margin_response)

            #  get the index price from clearing house vars and use that to
            # recalculate unrealised pnl and notional position
            positions = [dict(position) for position in margin_response['positions']]
            for position in positions:
                index_price = clearing_house_vars.underlying_prices[position['market']]
                index_price = int_to_scaled_float(index_price, 6)
                open_notional = float(position['openNotional'])
                size = float(position['size'])
                margin = float(margin_response['margin'])
                notional_position = abs(size * index_price)
                uPnL = notional_position - open_notional if size > 0 else open_notional - notional_position

                position['notionalPosition'] = round(notional_position, 6)
                position['unrealisedProfit'] = round(uPnL, 6)

            margin_response['positions'] = [AttributeDict(position) for position in positions]
            margin_response = AttributeDict(margin_response)

            return GetPositionsResponse(**margin_response)
        except ValueError as e:
            if len(e.args) > 0 and e.args[0].get('message', '') == "trader not found":
                raise TraderNotFound()
            else:
                raise e

    async def get_order_book_depth(self, market: int) -> OrderBookDepthResponse:
        response = await self._get_order_book_depth(market)
        return OrderBookDepthResponse(**response)

    async def get_clearing_house_vars(self, market: int) -> OrderBookDepthResponse:
        response = await self._get_order_book_depth(market)
        return response

    async def get_open_orders(self, trader:Address, market: int=None) -> List[OpenOrder]:
        if market is None:
            market_str = ""
        else:
            market_str = str(market)

        open_orders = []
        response = await self._get_open_orders(trader, market_str)
        for order in response["Orders"]:
            open_orders.append(OpenOrder(**order))
        return open_orders

    async def place_signed_orders(self, signed_orders) -> Dict:
        response = await self._place_signed_orders(signed_orders)
        orders = response.get('orders', [])
        response = []
        for order in orders:
            order_response = {
                'success': order.get('success', False),
                'order_id': order.get('orderId', ''),
            }
            if order.get('error', ''):
                order_response['error'] = order['error']

            response.append(order_response)
        return response

    async def wait_for_transaction_status(self, transaction_hash: HexBytes, timeout: float = 120, poll_latency: float = 0.1) -> Dict:
        async def _wait_for_status_with_timeout(
            _tx_hash: _Hash32, _poll_latency: float
        ) -> Dict:
            while True:
                try:
                    tx_status = await self._get_transaction_status(_tx_hash.hex())
                except:
                    tx_status = None
                if tx_status is None or tx_status['status'] == "NOT_FOUND":
                    await asyncio.sleep(_poll_latency)
                    continue
                else:
                    break
            return tx_status

        try:
            return await asyncio.wait_for(
                _wait_for_status_with_timeout(transaction_hash, poll_latency),
                timeout=timeout,
            )
        except asyncio.TimeoutError:
            raise TimeExhausted(
                f"Transaction {HexBytes(transaction_hash) !r} is not in the chain "
                f"after {timeout} seconds"
            )


class HubblenetWeb3(AsyncWeb3):
    eth: HubblenetEth


def get_web3_modules() -> Dict:
    modules = get_async_default_modules()
    modules["eth"] = HubblenetEth
    return modules


def get_async_web3_client() -> HubblenetWeb3:
    global async_web3_client
    if not async_web3_client:
        rpc_endpoint = get_rpc_endpoint()

        async_web3_client = HubblenetWeb3(AsyncHTTPProvider(rpc_endpoint), modules=get_web3_modules())
        async_web3_client.middleware_onion.inject(async_geth_poa_middleware, layer=0)

        # cache frequent eth_chainId calls
        async_web3_client.middleware_onion.add(_async_simple_cache_middleware, name="cache")

        # async_web3_client.eth.set_gas_price_strategy(rpc_gas_price_strategy)

    return async_web3_client


def get_sync_web3_client() -> Web3:
    global sync_web3_client
    if not sync_web3_client:
        rpc_endpoint = get_rpc_endpoint()

        sync_web3_client = Web3(HTTPProvider(rpc_endpoint))
        sync_web3_client.middleware_onion.inject(geth_poa_middleware, layer=0)

        sync_web3_client.middleware_onion.add(construct_simple_cache_middleware(), name="cache")

    return sync_web3_client


def get_rpc_endpoint() -> str:
    rpc_endpoint = os.getenv("HUBBLE_RPC")
    if not rpc_endpoint:
        raise ValueError("HUBBLE_RPC environment variable not set")
    return rpc_endpoint


def get_websocket_endpoint() -> str:
    ws_rpc_endpoint = os.getenv("HUBBLE_WS_RPC")
    if not ws_rpc_endpoint:
        raise ValueError("HUBBLE_WS_RPC environment variable not set")
    return ws_rpc_endpoint



import os

from hubble_exchange.constants import init_config

env = os.getenv("HUBBLE_ENV") or "mainnet"
config_module = __import__(f"hubble_exchange.config.{env}", fromlist=['*'])
init_config(config_module)


from hubble_exchange.client import HubbleClient
from hubble_exchange.models import (ConfirmationMode, GetPositionsResponse,
                                    IOCOrder, LimitOrder,
                                    OrderBookDepthResponse,
                                    OrderBookDepthUpdateResponse,
                                    OrderStatusResponse, SignedOrder)
from hubble_exchange.order_book import TransactionMode
